'A* pathfinding demo with terrain bias
'Mouse [Left-Click] to mark the target position
'Mouse [Right-click] to mark the start position
'Mouse [Middle-click] to clear a node
'Mouse [Middle-click]+[Left-Shift] to create a wall
'Press [R] to generate a chaotic random map
'Press [T] to toggle the algorithm ignoring terrain bias calculations
'Press [B] to toggle terrain bias map rendering
'Press [G] to clear the entire map to grass
'Press [F1] to "blur" the terrain bias map (experimental)
'This 'main' isn't important... it's just a way to have fun while experimenting
'The meat and potatoes is in aStar.bas/bi
'Dr_D

#include once "string.bi"
#include once "fbgfx.bi"
#include once "2dMathf.bi"
#include once "world.bi"
#include once "aStar.bi"
#include once "fbpng.bi"

type circ

	p as vec2i
	rad as integer
	bType as BlockType
	isFilled as integer

end type

declare function main( byval argc as integer, byval argv as zstring ptr ptr ) as integer
declare function random_map( map as worldstruct, dude as entitystruct ) as integer

end main( __fb_argc__, __fb_argv__ )


function main( byval argc as integer, byval argv as zstring ptr ptr ) as integer

	randomize timer

	const scr_w = 1280
	const scr_h = 1024

	screenres scr_w, scr_h, 32,, FB.GFX_HIGH_PRIORITY or FB.GFX_ALPHA_PRIMITIVES

	dim as WorldStruct map
	dim as EntityStruct dude
	dim as NavGrid myGrid

	dim as integer mouseX, mouseY, mouseB
	dim as integer mouseNodeX, mouseNodeY
	dim as integer pCnt
	dim as integer foundpath
	dim as integer showBias = false

	dim as double deltaTime, curTime, startCalc, finCalc, keyTime, startPathTime, finPathTime

	dim as vec2i rightClickPosition, leftClickPosition, midClickPosition

	dim as string statusMessage
	dim as string terType(4) =>{"WALL", "GRASS", "SAND", "MUD", "WATER"}

	redim map.image(4)

	map.image(0) = png_load("res/textures/terrain/wall1.png", PNG_TARGET_FBNEW)
	map.image(1) = png_load("res/textures/terrain/grass1.png", PNG_TARGET_FBNEW)
	map.image(2) = png_load("res/textures/terrain/sand1.png", PNG_TARGET_FBNEW)
	map.image(3) = png_load("res/textures/terrain/mud1.png", PNG_TARGET_FBNEW)
	map.image(4) = png_load("res/textures/terrain/water1.png", PNG_TARGET_FBNEW)


	random_map(map, dude)
	myGrid.update_node_map(map)




	curTime = timer

	do

		deltaTime = timer-curTime
		curTime = timer

		getmouse(mouseX, mouseY,,mouseB)
		mouseNodeX = cint(mouseX\SCALE)
		mouseNodeY = cint(mouseY\SCALE)

		if mouseNodeX < 0 then
			mouseNodeX = 0
		end if

		if mouseNodeY < 0 then
			mouseNodeY = 0
		end if

		if mouseNodeX > ubound(map.tile, 1) then
			mouseNodeX = ubound(map.tile, 1)
		end if

		if mouseNodeY > ubound(map.tile, 2) then
			mouseNodeY = ubound(map.tile, 2)
		end if

		if multikey(FB.SC_R) then

			random_map(map, dude)

			myGrid.update_node_map( map )

			pCnt = 0

			foundPath = 0

		end if

		if multikey(FB.SC_B) then

			if timer >= keyTime then

				keyTime = timer+.25

				showBias = not showBias

			end if

		end if

		if multikey(FB.SC_T) then

			if timer >= keyTime then

				keyTime = timer+.25

				myGrid.useTerrainBias = not myGrid.useTerrainBias

				myGrid.update_node_map( map )

			end if

		end if


		if multikey(FB.SC_F1) then

			if timer >= keyTime then

				keyTime = timer+.25

				myGrid.blend_bias_map()

			end if

		end if


		if multikey(FB.SC_G) then

			if timer >= keyTime then

				keyTime = timer+.25

				for y as integer = 0 to ubound(map.tile,2)

					for x as integer = 0 to ubound(map.tile,1)

						map.tile(x,y).bType = GRASS

						map.tile(x,y).id = 1

					next

				next

				myGrid.update_node_map( map )

			end if

		end if



		if (mouseB and FB.BUTTON_LEFT) and mouseX>-1 then

			if timer >= keyTime then

				keyTime = timer+.25

				leftClickPosition.x = mouseNodeX
				leftClickPosition.y = mouseNodeY

				dim as vec2i sPos
				sPos.x = cint(dude.p.x/SCALE)
				sPos.y = cint(dude.p.y/SCALE)

				startCalc = timer

				select case as const myGrid.find_path(dude, sPos, leftClickPosition)

					case status_state.nBadStart

						foundpath = false
						statusMessage = "Bad starting position!"

					case status_state.nBadEnd

						foundpath = false
						statusMessage = "Bad target position!"

					case status_state.nSuccess

						foundpath = true
						pCnt = 0
						statusMessage = "Found a connecting path!"

					case status_state.nIncomplete

						foundpath = false
						statusMessage = "Cannot complete a path here!"

				end select

				finCalc = timer-startCalc
				startPathTime = timer
				
			end if

		end if


		if (mouseB and FB.BUTTON_MIDDLE) and mouseX>-1 then

			midClickPosition.x = mouseNodeX
			midClickPosition.y = mouseNodeY

			if multikey(FB.SC_LSHIFT) then

				map.tile(midClickPosition.x, midClickPosition.y).bType = WALL
				map.tile(midClickPosition.x, midClickPosition.y).id = 0

			else

				map.tile(midClickPosition.x, midClickPosition.y).bType = GRASS
				map.tile(midClickPosition.x, midClickPosition.y).id = 1

			end if

			myGrid.update_node_map( map )

		end if


		if (mouseB and FB.BUTTON_RIGHT) and mouseX>-1 then

			rightClickPosition.x = mouseNodeX
			rightClickPosition.y = mouseNodeY

			dude.p.x = rightClickPosition.x * SCALE
			dude.p.y = rightClickPosition.y * SCALE
			
			pCnt = 0
			
			foundPath = 0
			
		end if



		if foundpath then

			dim as integer dudeX = cint(dude.p.x / SCALE)
			dim as integer dudeY = cint(dude.p.y / SCALE)

			dim as single mSpeed = 100 - myGrid.gNode(dudeX, dudeY).c * 1.5

			var pVector = vec2f(dude.path.p(pCnt).x * SCALE, dude.path.p(pCnt).y * SCALE)

			dim as vec2f dVector = pVector - vec2f(cint(dude.p.x), cint(dude.p.y))

			dim as single pMag = dVector.magnitude()

			dVector.Normalize()

			dude.p += dVector * mSpeed * deltaTime

			dude.fcntr += (mSpeed / 4.85) * deltaTime

			if int(dude.fcntr) > 15 then dude.fcntr = 0

			dVector = vec2f(dVector.x \ 1, dVector.y \ 1)

			dim as integer id
			for n as integer = 0 to 7

				if dVector.x = dude.dr(n).x and dVector.y = dude.dr(n).y then

					id = n

				end if

			next

			dude.direction = id

			if pMag <= SCALE/2 then

				pCnt+=1

				if pCnt > ubound(dude.path.p) then

					pCnt = ubound(dude.path.p)

					dude.p.x = dude.path.p(pCnt).x * SCALE
					dude.p.y = dude.path.p(pCnt).y * SCALE

					foundpath = false
					
					finPathTime = timer - startPathTime

					statusMessage = "Left-Click a block to create a path..."

				end if

			end if

		end if

		screenlock

		cls

		drawmap(map)

		if showBias then

			for y as integer = 0 to ubound(map.tile,2)

				for x as integer = 0 to ubound(map.tile,1)

					with myGrid.gNode(x, y)

						dim as uinteger cCol = rgba(.c*3, .c*3, .c*3, 255)

						if .blocked then .c = rgba(255, 0, 0, 255)

						line(x * SCALE, y * SCALE)-step(SCALE - 1, SCALE - 1), rgba(cCol, cCol, cCol, 250), bf

					end with

				next

			next

		end if

		for i as integer = 0 to ubound(dude.path.p)

			dim as uinteger cRad = 4, cColor = rgba(255, 255, 255, 96)

			if pCnt = i then

				cRad = 7

				cColor = rgba(255, 255, 255, 160)

			elseif pCnt>i then

				cRad = 2

				cColor = rgba(0, 0, 0, 128)

			end if

			if i = 0 or i = ubound(dude.path.p) then

				cColor = rgba(255, 255, 0, 160)

			end if

			circle(dude.path.p(i).x * SCALE + SCALE / 2, dude.path.p(i).y * SCALE + SCALE / 2), cRad+2, rgba(0,0,0,32),,,, f
			circle(dude.path.p(i).x * SCALE + SCALE / 2, dude.path.p(i).y * SCALE + SCALE / 2), cRad, cColor,,,, f

		next

		drawcharacter(dude)



		if myGrid.gNode(mouseNodeX, mouseNodeY).blocked then

			line(mouseNodeX * SCALE, mouseNodeY * SCALE) - step(SCALE, SCALE), rgb(255, 255, 255)
			line(mouseNodeX * SCALE, mouseNodeY * SCALE + SCALE) - step(SCALE, -SCALE), rgb(255, 255, 255)

		else

			line(mouseNodeX * SCALE + 1, mouseNodeY * SCALE + 1) - step(SCALE - 2, SCALE - 2), rgb(255, 255, 255), b

		end if


		draw string (0, scr_h-9), str(ubound(dude.path.p)) & " nodes generated in " & format(finCalc, "##.####") & " seconds"
		draw string (0, scr_h-18), "Character is at X " & cint(dude.p.x\Scale) & ", Y " & cint(dude.p.y\Scale) & " on " & terType(map.tile(cint(dude.p.x\Scale), cint(dude.p.y\Scale)).id)
		
		dim as string biasString = "OFF"
		if myGrid.useTerrainBias then biasString = "ON"
		draw string (0, scr_h-27), "Terrain Bias = " & biasString
		draw string (0, scr_h-36), "Path traversed in " & format( finPathTime, "##.####") & " seconds", rgb(255, 255, 255)
		draw string (0, scr_h-45), statusMessage, rgb(255, 255, 0)
		
		
		screensync
		screenunlock

		sleep(3, 1)

	loop until multikey(FB.SC_ESCAPE)


	for i as integer = 0 to ubound(map.image)
		png_destroy(map.image(i))
	next

	return argc

end function


function random_map( map as worldstruct, dude as entitystruct ) as integer

	map.size = type(79, 60)
	redim map.tile(map.size.x, map.size.y)



	dim c(300) as circ
	reMap:

	for i as integer = 0 to ubound(c)

		c(i).p = type(int(rnd*ubound(map.tile, 1))+1,int(rnd*ubound(map.tile, 2))+1)

		c(i).rad = 2+rnd*6

		c(i).bType = cast (BlockType, int(rnd*5) )

		c(i).isFilled = int(rnd*10)

	next


	for y as integer = 0 to ubound(map.tile, 2)

		for x as integer = 0 to ubound(map.tile, 1)

			map.tile(x,y).bType = GRASS
			map.tile(x,y).id = 1


			for i as integer = 0 to ubound(c)

				dim as integer dist = sqr((c(i).p.y-y)^2 + (c(i).p.x-x)^2)

				if dist = c(i).rad then

					map.tile(x,y).bType = c(i).bType

					map.tile(x,y).id = c(i).bType

				end if

				if c(i).isFilled > 0 then

					if dist < c(i).rad then

						map.tile(x,y).bType = c(i).bType

						map.tile(x,y).id = c(i).bType

					end if

				end if

			next


			if x = 0 or y = 0 or x = ubound(map.tile,1) or y = ubound(map.tile,2) then

				map.tile(x,y).bType = WALL

				map.tile(x,y).id = 0

			end if

		next

	next


	do

		var x = int(rnd*map.size.x+1)
		var y = int(rnd*map.size.y+1)

		if map.tile(x,y).bType = WALL then
		else
			dude.p = vec2f(x*SCALE,y*SCALE)
			exit do
		end if


	loop


	return true

end function

